@@ -2,17 +2,22 @@ import logging
 from os import mkdir
 from os.path import join
 from shutil import rmtree
+from typing import Any, Iterator, List, Optional, Tuple, Union
 
-from hdx.utilities.downloader import DownloadError
-from hdx.utilities.loader import load_file_to_str, load_json, load_yaml
-from hdx.utilities.saver import save_json, save_str_to_file, save_yaml
+from hdx.utilities.base_downloader import BaseDownload, DownloadError
+from hdx.utilities.downloader import Download
+from hdx.utilities.loader import load_json, load_text, load_yaml
+from hdx.utilities.path import get_filename_from_url
+from hdx.utilities.saver import save_json, save_text, save_yaml
+from hdx.utilities.typehint import ListDict, ListTuple
 
 logger = logging.getLogger(__name__)
 
 
-class Retrieve:
-    """Retrieve class which takes in a Download object and can either download, download and save or use previously
-    downloaded and saved data. It also allows the use of a static fallback when downloading fails.
+class Retrieve(BaseDownload):
+    """Retrieve class which takes in a Download object and can either download, download
+    and save or use previously downloaded and saved data. It also allows the use of a
+    static fallback when downloading fails.
 
     Args:
         downloader (Download): Download object
@@ -23,8 +28,16 @@ class Retrieve:
         use_saved (bool): Whether to use saved data. Defaults to False.
     """
 
+    retrievers = dict()
+
     def __init__(
-        self, downloader, fallback_dir, saved_dir, temp_dir, save=False, use_saved=False
+        self,
+        downloader: Download,
+        fallback_dir: str,
+        saved_dir: str,
+        temp_dir: str,
+        save: bool = False,
+        use_saved: bool = False,
     ):
         self.downloader = downloader
         self.fallback_dir = fallback_dir
@@ -55,12 +68,19 @@ class Retrieve:
             return f"{url[:100]}..."
         return url
 
-    def retrieve_file(self, url, filename, logstr=None, fallback=False, **kwargs):
+    def download_file(
+        self,
+        url: str,
+        filename: Optional[str] = None,
+        logstr: Optional[str] = None,
+        fallback: bool = False,
+        **kwargs: Any,
+    ) -> str:
         """Retrieve file
 
         Args:
             url (str): URL to download
-            filename (str): Filename to use for saved file
+            filename (Optional[str]): Filename of saved file. Defaults to getting from url.
             logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
             fallback (bool): Whether to use static fallback if download fails. Defaults to False.
             **kwargs: Parameters to pass to download_file call
@@ -69,6 +89,8 @@ class Retrieve:
             str: Path to downloaded file
 
         """
+        if not filename:
+            filename = get_filename_from_url(url)
         if not logstr:
             logstr = filename
         if self.save:
@@ -85,7 +107,9 @@ class Retrieve:
                 logger.info(
                     f"Downloading {logstr} from {self.get_url_logstr(url)} into {output_path}"
                 )
-                return self.downloader.download_file(url, path=output_path, **kwargs)
+                return self.downloader.download_file(
+                    url, path=output_path, **kwargs
+                )
             except DownloadError:
                 if not fallback:
                     raise
@@ -95,34 +119,44 @@ class Retrieve:
                 )
                 return fallback_path
 
-    def retrieve_text(self, url, filename, logstr=None, fallback=False, **kwargs):
-        """Retrieve text
+    def download_text(
+        self,
+        url: str,
+        filename: Optional[str] = None,
+        logstr: Optional[str] = None,
+        fallback: bool = False,
+        **kwargs: Any,
+    ) -> str:
+        """Download text
 
         Args:
             url (str): URL to download
-            filename (str): Filename to use for saved file
+            filename (Optional[str]): Filename of saved file. Defaults to getting from url.
             logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
             fallback (bool): Whether to use static fallback if download fails. Defaults to False.
             **kwargs: Parameters to pass to download call
 
         Returns:
-            Union[Dict,List]: The text from the file
+            str: The text from the file
 
         """
+        if not filename:
+            filename = get_filename_from_url(url)
         if not logstr:
             logstr = filename
         saved_path = join(self.saved_dir, filename)
         if self.use_saved:
             logger.info(f"Using saved {logstr} in {saved_path}")
-            text = load_file_to_str(saved_path)
+            text = load_text(saved_path)
         else:
             try:
-                logger.info(f"Downloading {logstr} from {self.get_url_logstr(url)}")
-                self.downloader.download(url, **kwargs)
-                text = self.downloader.get_text()
+                logger.info(
+                    f"Downloading {logstr} from {self.get_url_logstr(url)}"
+                )
+                text = self.downloader.download_text(url, **kwargs)
                 if self.save:
                     logger.info(f"Saving {logstr} in {saved_path}")
-                    save_str_to_file(text, saved_path)
+                    save_text(text, saved_path)
             except DownloadError:
                 if not fallback:
                     raise
@@ -130,23 +164,32 @@ class Retrieve:
                 logger.exception(
                     f"{logstr} download failed, using static data {fallback_path}!"
                 )
-                text = load_file_to_str(fallback_path)
+                text = load_text(fallback_path)
         return text
 
-    def retrieve_yaml(self, url, filename, logstr=None, fallback=False, **kwargs):
+    def download_yaml(
+        self,
+        url: str,
+        filename: Optional[str] = None,
+        logstr: Optional[str] = None,
+        fallback: bool = False,
+        **kwargs: Any,
+    ) -> Any:
         """Retrieve YAML
 
         Args:
             url (str): URL to download
-            filename (str): Filename to use for saved file
+            filename (Optional[str]): Filename of saved file. Defaults to getting from url.
             logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
             fallback (bool): Whether to use static fallback if download fails. Defaults to False.
             **kwargs: Parameters to pass to download call
 
         Returns:
-            Union[Dict,List]: The data from the YAML file
+            Any: The data from the YAML file
 
         """
+        if not filename:
+            filename = get_filename_from_url(url)
         if not logstr:
             logstr = filename
         saved_path = join(self.saved_dir, filename)
@@ -155,9 +198,10 @@ class Retrieve:
             ryaml = load_yaml(saved_path)
         else:
             try:
-                logger.info(f"Downloading {logstr} from {self.get_url_logstr(url)}")
-                self.downloader.download(url, **kwargs)
-                ryaml = self.downloader.get_yaml()
+                logger.info(
+                    f"Downloading {logstr} from {self.get_url_logstr(url)}"
+                )
+                ryaml = self.downloader.download_yaml(url, **kwargs)
                 if self.save:
                     logger.info(f"Saving {logstr} in {saved_path}")
                     save_yaml(ryaml, saved_path)
@@ -171,20 +215,29 @@ class Retrieve:
                 ryaml = load_yaml(fallback_path)
         return ryaml
 
-    def retrieve_json(self, url, filename, logstr=None, fallback=False, **kwargs):
+    def download_json(
+        self,
+        url: str,
+        filename: Optional[str] = None,
+        logstr: Optional[str] = None,
+        fallback: bool = False,
+        **kwargs: Any,
+    ) -> Any:
         """Retrieve JSON
 
         Args:
             url (str): URL to download
-            filename (str): Filename to use for saved file
+            filename (Optional[str]): Filename of saved file. Defaults to getting from url.
             logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
             fallback (bool): Whether to use static fallback if download fails. Defaults to False.
             **kwargs: Parameters to pass to download call
 
         Returns:
-            Union[Dict,List]: The data from the JSON file
+            Any: The data from the JSON file
 
         """
+        if not filename:
+            filename = get_filename_from_url(url)
         if not logstr:
             logstr = filename
         saved_path = join(self.saved_dir, filename)
@@ -193,9 +246,10 @@ class Retrieve:
             rjson = load_json(saved_path)
         else:
             try:
-                logger.info(f"Downloading {logstr} from {self.get_url_logstr(url)}")
-                self.downloader.download(url, **kwargs)
-                rjson = self.downloader.get_json()
+                logger.info(
+                    f"Downloading {logstr} from {self.get_url_logstr(url)}"
+                )
+                rjson = self.downloader.download_json(url, **kwargs)
                 if self.save:
                     logger.info(f"Saving {logstr} in {saved_path}")
                     save_json(rjson, saved_path)
@@ -208,3 +262,82 @@ class Retrieve:
                 )
                 rjson = load_json(fallback_path)
         return rjson
+
+    def get_tabular_rows(
+        self,
+        url: str,
+        headers: Union[int, ListTuple[int], ListTuple[str]] = 1,
+        dict_form: bool = False,
+        filename: Optional[str] = None,
+        logstr: Optional[str] = None,
+        fallback: bool = False,
+        **kwargs: Any,
+    ) -> Tuple[List[str], Iterator[ListDict]]:
+        """Returns header of tabular file pointed to by url and an iterator where each
+        row is returned as a list or dictionary depending on the dict_rows argument.
+        The headers argument is either a row number or list of row numbers (in case of
+        multi-line headers) to be considered as headers (rows start counting at 1), or
+        the actual headers defined as a list of strings. It defaults to 1.
+        The dict_form arguments specifies if each row should be returned as a dictionary
+        or a list, defaulting to a list.
+
+        Args:
+            url (str): URL or path to read from
+            headers (Union[int, ListTuple[int], ListTuple[str]]): Number of row(s) containing headers or list of headers. Defaults to 1.
+            dict_form (bool): Return dict or list for each row. Defaults to False (list)
+            filename (Optional[str]): Filename of saved file. Defaults to getting from url.
+            logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
+            fallback (bool): Whether to use static fallback if download fails. Defaults to False.
+            **kwargs: Parameters to pass to download call
+
+        Returns:
+            Tuple[List[str],Iterator[ListDict]]: Tuple (headers, iterator where each row is a list or dictionary)
+
+        """
+        path = self.download_file(url, filename, logstr, fallback, **kwargs)
+        return self.downloader.get_tabular_rows(
+            path, headers, dict_form, **kwargs
+        )
+
+    @classmethod
+    def generate_retrievers(
+        cls,
+        fallback_dir: str,
+        saved_dir: str,
+        temp_dir: str,
+        save: bool = False,
+        use_saved: bool = False,
+    ) -> None:
+        """Generate retrievers. Retrievers are generated from downloaders so
+        Download.generate_downloaders() needs to have been called first. Each retriever
+        can either download, download and save or use previously downloaded and saved
+        data. It also allows the use of a static fallback when downloading fails.
+
+        Args:
+            fallback_dir (str): Directory containing static fallback data
+            saved_dir (str): Directory to save or load downloaded data
+            temp_dir (str): Temporary directory for when data is not needed after downloading
+            save (bool): Whether to save downloaded data. Defaults to False.
+            use_saved (bool): Whether to use saved data. Defaults to False.
+
+        Returns:
+            None
+        """
+        cls.retrievers = dict()
+        for name, downloader in Download.downloaders.items():
+            cls.retrievers[name] = cls(
+                downloader, fallback_dir, saved_dir, temp_dir, save, use_saved
+            )
+
+    @classmethod
+    def get_retriever(cls, name: Optional[str] = None) -> "Retrieve":
+        """Get a generated retriever given a name. If name is not supplied, the default
+        one will be returned.
+
+        Args:
+            name (Optional[str]): Name of retriever. Defaults to None (get default).
+
+        Returns:
+            Retriever: Retriever object
+        """
+        return cls.retrievers.get(name, cls.retrievers["default"])
